#define GL_DO_NOT_WARN_IF_MULTI_GL_VERSION_HEADERS_INCLUDED

#include <Cocoa/Cocoa.h>
#import <GLKit/GLKit.h>
#include <servers/bootstrap.h>
#import "MultiGPUMig.h"
#import "MultiGPUMigServer.h"

#include "MacIOSurfaceManager.h"


// On OS 10.13 or later, use MachO comunication and IOSurfaceBuffer to
// display the graphics output of our child graphics apps in our window.

// Code adapted from Apple Developer Tech Support Sample Code MutiGPUIOSurface:
// <https://developer.apple.com/library/content/samplecode/MultiGPUIOSurface>

#define NUM_IOSURFACE_BUFFERS 2

@interface MachServerController : NSObject <NSMachPortDelegate>
{
	NSMachPort *serverPort;
	NSMachPort *localPort;

	uint32_t serverPortName;
	uint32_t localPortName;

	NSMachPort *clientPort[16];
	uint32_t clientPortNames[16];
	uint32_t clientPortCount;
}
- (MachServerController *)init;
- (kern_return_t)checkInClient:(mach_port_t)client_port index:(int32_t *)client_index;
- (void)portDied:(NSNotification *)notification;
- (void)sendIOSurfaceMachPortToClients: (uint32_t)index withMachPort:(mach_port_t) iosurface_port;
- (bool)activeClients;
@end

static MachServerController *myserverController;

static uint32_t currentFrameIndex;
static GLsizei w, h;

static IOSurfaceRef ioSurfaceBuffers[NUM_IOSURFACE_BUFFERS];
static mach_port_t ioSurfaceMachPorts[NUM_IOSURFACE_BUFFERS];
static GLuint textureNames[NUM_IOSURFACE_BUFFERS];
static GLuint fboNames[NUM_IOSURFACE_BUFFERS];
static GLuint depthBufferName;

@implementation MachServerController
- (MachServerController *)init
{
	[[NSNotificationCenter defaultCenter] addObserver:self
	 selector:@selector(portDied:) name:NSPortDidBecomeInvalidNotification object:nil];

	mach_port_t servicePortNum = MACH_PORT_NULL;
	kern_return_t machErr;
	char *portName = "edu.berkeley.boincsaver";

	machErr = bootstrap_check_in(bootstrap_port, portName, &servicePortNum);
	if (machErr != KERN_SUCCESS) {
		[NSApp terminate:self];
	}
	serverPort = (NSMachPort*)[NSMachPort portWithMachPort:servicePortNum];

	// Create a local dummy reply port to use with the mig reply stuff
	localPort = [[NSMachPort alloc] init];

	// Retrieve raw mach port names.
	serverPortName = [serverPort machPort];
	localPortName  = [localPort machPort];

	[serverPort setDelegate:self];
	[serverPort scheduleInRunLoop:[NSRunLoop currentRunLoop] forMode:NSRunLoopCommonModes];

	return self;
}

- (void)portDied:(NSNotification *)notification
{
	NSPort *port = [notification object];
	if(port == serverPort)
	{
		[NSApp terminate:self];
	}
	else
	{
		int i;
		for(i = 0; i < clientPortCount+1; i++)
		{
			if([clientPort[i] isEqual:port])
			{
				[clientPort[i] release];
				clientPort[i] = nil;
				clientPortNames[i] = 0;
			}
		}
	}
}

- (void)handleMachMessage:(void *)msg
{
	union __ReplyUnion___MGCMGSServer_subsystem reply;

	mach_msg_header_t *reply_header = (void *)&reply;
	kern_return_t kr;

	if(MGSServer_server(msg, reply_header) && reply_header->msgh_remote_port != MACH_PORT_NULL)
	{
		kr = mach_msg(reply_header, MACH_SEND_MSG, reply_header->msgh_size, 0, MACH_PORT_NULL,
		              0, MACH_PORT_NULL);
		if(kr != 0)
			[NSApp terminate:nil];
	}
}

- (kern_return_t)checkInClient:(mach_port_t)client_port index:(int32_t *)client_index
{
	clientPortCount++;                          // clients always start at index 1
	clientPortNames[clientPortCount] = client_port;
	clientPort[clientPortCount] = [[NSMachPort alloc] initWithMachPort:client_port];

	*client_index = clientPortCount;
	return 0;
}

kern_return_t _MGSCheckinClient(mach_port_t server_port, mach_port_t client_port,
                                int32_t *client_index)
{
	return [myserverController checkInClient:client_port index:client_index];
}

// For the MachO server, this is a no-op
kern_return_t _MGSDisplayFrame(mach_port_t server_port, int32_t frame_index, uint32_t iosurface_port)
{
	return 0;
}

- (void)sendIOSurfaceMachPortToClients:(uint32_t)index withMachPort:(mach_port_t)iosurface_port
{
	int i;
	for(i = 0; i < clientPortCount+1; i++)
	{
		if(clientPortNames[i])
		{
			_MGCDisplayFrame(clientPortNames[i], index, iosurface_port);
		}
	}
}

- (bool)activeClients
{
	return clientPortCount > 0 ? true : false;
}
@end

void initMacIOSurface() {
	int viewportRect[4];
	glGetIntegerv(GL_VIEWPORT, (GLint*)viewportRect);
	w = viewportRect[2];
	h = viewportRect[3];

	if (!myserverController) {
		myserverController = [[[MachServerController alloc] init] retain];
	}

	if (!ioSurfaceBuffers[0]) {
		// Set up all of our iosurface buffers
		for(int i = 0; i < NUM_IOSURFACE_BUFFERS; i++) {
			ioSurfaceBuffers[i] = IOSurfaceCreate((CFDictionaryRef) @{
				(id)kIOSurfaceWidth: [NSNumber numberWithInt: w],
				(id)kIOSurfaceHeight: [NSNumber numberWithInt: h],
				(id)kIOSurfaceBytesPerElement: @4
				});
			ioSurfaceMachPorts[i] = IOSurfaceCreateMachPort(ioSurfaceBuffers[i]);
		}
	}
}

void renderMacIOSurfacePre() {
	GLuint name, namef;

	if(!myserverController || ![myserverController activeClients]) {
		return;
	}

	if(!textureNames[currentFrameIndex]) {
		NSOpenGLContext * myContext = [ NSOpenGLContext currentContext ];
		CGLContextObj cgl_ctx = (CGLContextObj)[myContext CGLContextObj];

		glGenTextures(1, &name);

		glBindTexture(GL_TEXTURE_RECTANGLE, name);
		// At the moment, CGLTexImageIOSurface2D requires the GL_TEXTURE_RECTANGLE target
		CGLTexImageIOSurface2D(cgl_ctx, GL_TEXTURE_RECTANGLE, GL_RGBA, w, h, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV,
		                       ioSurfaceBuffers[currentFrameIndex], 0);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

		// Generate an FBO and bind the texture to it as a render target.
		glBindTexture(GL_TEXTURE_RECTANGLE, 0);

		glGenFramebuffers(1, &namef);
		glBindFramebuffer(GL_FRAMEBUFFER, namef);
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_RECTANGLE, name, 0);

		if(!depthBufferName) {
			glGenRenderbuffers(1, &depthBufferName);
			glRenderbufferStorage(GL_TEXTURE_RECTANGLE, GL_DEPTH, w, h);
		}

		glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_RECTANGLE, depthBufferName);

		fboNames[currentFrameIndex] = namef;
		textureNames[currentFrameIndex] = name;
	}

	glBindFramebuffer(GL_FRAMEBUFFER, fboNames[currentFrameIndex]);
}

bool renderMacIOSurfacePost() {

	if(!myserverController || ![myserverController activeClients]) {
		return false;
	}

	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glFlush();

	[myserverController sendIOSurfaceMachPortToClients: currentFrameIndex
	 withMachPort:ioSurfaceMachPorts[currentFrameIndex]];

	currentFrameIndex = (currentFrameIndex + 1) % NUM_IOSURFACE_BUFFERS;

	return true;
}
